data List (A : Set) : Set where
  []  : List A
  _∷_ : A → List A → List A

infixr 5 _∷_

open import Data.Nat
open import Relation.Binary.PropositionalEquality

length : {A : Set} → List A → ℕ
length [] = 0
length (x ∷ L) = suc (length L)

concat : {A : Set} → List A → List A → List A
concat [] L' = L'
concat (x ∷ L) L' = x ∷ (concat L L')

concat-length :  {A : Set} → (L : List A) → (L' :  List A) → length (concat L L') ≡ length L + length L'
concat-length [] L' = refl
concat-length (x ∷ L) L' = cong suc (concat-length L L')

cong-push :  {A : Set} → (x : A) → (L L' : List A) → L ≡ L' → x ∷ L ≡ x ∷ L'
cong-push x L .L refl = refl

concat-assoc :  {A : Set} → (L M N : List A) → concat L (concat M N) ≡ concat (concat L M) N
concat-assoc [] M N = refl
concat-assoc (x ∷ L) M N = cong-push x _ _ (concat-assoc  L M N)

snoc :  {A : Set} → A  → List A → List A
snoc x [] = x ∷ []
snoc x (y ∷ L) = y ∷ (snoc x L)

rev :  {A : Set} → List A → List A
rev [] = []
rev (x ∷ L) = snoc x (rev L)

snoc-length :  {A : Set} → (x : A) → (L : List A) → length (snoc x L) ≡ suc (length L)
snoc-length x [] = refl
snoc-length x (y ∷ L) = cong suc (snoc-length x L)

rev-length :  {A : Set} → (L : List A) → length (rev L) ≡ length L
rev-length [] = refl
rev-length (x ∷ L) = trans (snoc-length x (rev L)) (cong suc (rev-length L))

rev-snoc : {A : Set} → (x : A) → (L : List A) → rev (snoc x L) ≡ x ∷ (rev L)
rev-snoc x [] = refl
rev-snoc x (y ∷ L) = trans (cong (snoc y) (rev-snoc x L)) (cong-push x (snoc y (rev L)) (snoc y (rev L)) refl)

rev-involutive :  {A : Set} → (L : List A) → rev (rev L) ≡ L
rev-involutive [] = refl
rev-involutive (x ∷ L) = trans (rev-snoc x (rev L)) (cong-push x (rev (rev L)) L (rev-involutive L))

open import Data.Bool

filter : {A : Set} → (p : A → Bool) → (l : List A) → List A
filter p [] = []
filter p (x ∷ l) = if (p x) then (x ∷ (filter p l)) else (filter p l)

always-false : {A : Set} → A → Bool
always-false _  = false

always-true : {A : Set} → A → Bool
always-true _  = true

filter-false : {A : Set} → (l : List A) → filter always-false l ≡ []
filter-false [] = refl
filter-false (x ∷ l) = filter-false l

filter-true : {A : Set} → (l : List A) → filter always-true l ≡ l
filter-true [] = refl
filter-true (x ∷ l) = cong-push x (filter (λ _ → true) l) l (filter-true l)
