data ℕ : Set where
  zero : ℕ
  suc : ℕ → ℕ

_+_ : ℕ → ℕ → ℕ
zero + m = m
suc n + m = suc (n + m)

_*_ : ℕ → ℕ → ℕ
zero *  m = zero
suc n * m = (n *  m) + m

infix 21 _*_

data _≡_ {A : Set} (x : A) : (y : A) → Set where
  refl : x ≡ x

infix 4 _≡_

suc-≡ : {m n : ℕ} → (m ≡ n) → (suc m ≡ suc n)
suc-≡ refl = refl

O-plus-n : {n : ℕ} → zero + n ≡ n
O-plus-n = refl

n-plus-O : (n : ℕ) → n + zero ≡ n
n-plus-O zero = refl
n-plus-O (suc n) = suc-≡ (n-plus-O n)

plus-assoc : (m n p : ℕ) → (m + n) + p ≡ m + (n + p)
plus-assoc zero n p = refl
plus-assoc (suc m) n p = suc-≡ (plus-assoc m n p)

suc-assoc : (m n : ℕ) → suc (m + n) ≡ m + suc n
suc-assoc zero n = refl
suc-assoc (suc m) n = suc-≡ (suc-assoc m n)

open import Relation.Nullary

O-neq-suc : {n : ℕ} → ¬ zero ≡ suc n
O-neq-suc = λ ()

rec : (P : ℕ → Set) → P zero → ((n : ℕ) → P n → P (suc n)) →
  ((n : ℕ) -> P n)
rec P Pzero Psuc zero = Pzero
rec P Pzero Psuc (suc n) =  Psuc n (rec P Pzero Psuc n)

n-plus-O-rec : (n : ℕ) → n + zero ≡ n
n-plus-O-rec = rec _  refl λ n x → suc-≡ x
sym : {A : Set} {x y : A} → x ≡ y → y ≡ x
sym refl = refl

trans : {A : Set} {x y z : A} → x ≡ y → y ≡ z → x ≡ z
trans refl refl = refl

cong : {A B : Set} (f : A → B) {x y : A} → x ≡ y → f x ≡ f y
cong f refl = refl

suc-≡-cong : {m n : ℕ} → (m ≡ n) → (suc m ≡ suc n)
suc-≡-cong = cong suc

suc-inj : {m n : ℕ} → suc m ≡ suc n → m ≡ n
suc-inj refl = refl

open import Data.Sum renaming (_⊎_ to _∨_ ; inj₁ to left ; inj₂ to right)

nat-eq-dec : (m n : ℕ) → (m ≡ n) ∨ ¬ (m ≡ n)
nat-eq-dec zero zero = left refl
nat-eq-dec zero (suc n) = right O-neq-suc
nat-eq-dec (suc m) zero = right (λ ())
nat-eq-dec (suc m) (suc n) with (nat-eq-dec m n)
nat-eq-dec (suc m) (suc n) | left e = left (cong suc e)
nat-eq-dec (suc m) (suc n) | right f = right λ e → f (suc-inj e)

J : (A : Set) → (P : (x : A) → (y : A) → (p : x ≡ y) → Set) → (r : (x : A) → P x x refl) → (x : A) → (y : A) → (p : x ≡ y) → P x y p
J A P r x .x refl = r x

plus-com : (m n : ℕ) → m + n ≡ n +  m
plus-com zero n = sym (n-plus-O _)
plus-com (suc m) n = trans (suc-≡ (plus-com m n)) (suc-assoc n m)

plus-eq-left : (k m n : ℕ) → m ≡ n → k + m ≡ k + n
plus-eq-left zero = λ m n x → x
plus-eq-left (suc k) = λ m n x → suc-≡ (plus-eq-left  k m n x)


plus-eq-right : (k m n : ℕ) → m ≡ n → m + k ≡ n + k
plus-eq-right = λ k m n x → trans (plus-com m k) (trans (plus-eq-left k m n x) (plus-com k n))

mul-com : (m n : ℕ) → m * n ≡ n * m
mul-com zero zero = refl
mul-com zero (suc n) = trans (mul-com zero n) (sym (n-plus-O (n * zero)))
mul-com (suc m) zero =  trans  (n-plus-O (m * zero)) (mul-com m zero)
mul-com (suc m) (suc n) = trans (plus-eq-right (suc n) (m * suc n) ((n * m) + m)  (mul-com m (suc n)))
                          (trans (trans (sym (suc-assoc ((n * m) + m) n))
                          (trans (suc-≡
                          (trans (plus-assoc (n * m) m n)
                          (trans (trans (plus-eq-left (n * m) (m + n) (n + m) (plus-com m n))
                          (plus-eq-right (n + m) (n * m) (m * n) (mul-com n m)))
                          (sym (plus-assoc (m * n) n m)))))
                          (suc-assoc ((m * n) + n) m)))
                          (plus-eq-right (suc m) ((m * n) + n) (n * suc m) (mul-com (suc m) n)))


mul-distr : (m n p : ℕ) → (m + n) * p ≡ m * p + n * p
mul-distr zero n p = refl
mul-distr (suc m) n p = trans (plus-eq-right p ((m + n) * p) ((m * p) + (n * p)) (mul-distr m n p)) (trans (plus-assoc (m * p) (n * p) p)
                        (trans (plus-eq-left (m * p) ((n * p) + p) (p + (n * p)) (plus-com (n * p) p)) (sym(plus-assoc (m * p) p (n * p)))))


mul-assoc : (m n p : ℕ) → (m * n) * p ≡ m * (n * p)
mul-assoc zero n p = refl
mul-assoc (suc m) n p = trans (mul-distr (m * n) n p) (plus-eq-right (n * p) ((m * n) * p) (m * (n * p)) (mul-assoc m n p))

