open import Data.Nat

data Vec (A : Set) : ℕ → Set where
  []  : Vec A zero
  _∷_ : {n : ℕ} → A → Vec A n → Vec A (suc n)

head : {A : Set} → {n : ℕ} → Vec A (suc n) → A
head (x ∷ l) = x

tail : {A : Set} → {n : ℕ} → Vec A (suc n) → Vec A n
tail (x ∷ l) = l

concatenation : {A : Set} → {m n : ℕ} → Vec A m → Vec A n → Vec A (m + n)
concatenation [] l = l
concatenation (hd ∷ tl) l = hd ∷ (concatenation tl l)

sendtoend : {A : Set} → {m : ℕ} → Vec A m → A → Vec A (suc m)
sendtoend [] y = y ∷ []
sendtoend (x ∷ l) y = x ∷ (sendtoend l y)

reverse : {A : Set} → {m : ℕ} → Vec A m → Vec A m
reverse [] = []
reverse (x ∷ l) = sendtoend (reverse l) x

data Fin : ℕ → Set where
  zero : {n : ℕ} → Fin (suc n)
  suc  : {n : ℕ} (i : Fin n) → Fin (suc n)

access : {A : Set} → {m : ℕ} → Vec A m → Fin m → A
access (x ∷ l) zero = x
access (x ∷ l) (suc i) = access l i

open import Data.Product hiding (zip)

zip :  {A B : Set} → {m : ℕ} → Vec A m → Vec B m → Vec (A × B) m
zip [] [] = []
zip (x ∷ lA) (y ∷ lB) = {!(x , y) ∷ (zip lA lB)!}

open import Relation.Binary.PropositionalEquality

coe : {A B : Set} → A ≡ B → A → B
coe refl a = a

open import Data.Nat.Properties using (+-assoc)

vect-eq-1 :  {A : Set} → {m n : ℕ} → m ≡ n → Vec A m ≡ Vec A n
vect-eq-1 refl = refl

  
vect-eq-2 :  {A : Set} → {m n : ℕ} → {x : A} →  (lm : Vec A m) → (ln : Vec A n) →(e : m ≡ n) → coe (vect-eq-1 e) lm ≡ ln → coe (vect-eq-1 (cong suc e))( x ∷ lm) ≡ x ∷ ln
vect-eq-2 lm ln refl refl = refl

vect-eq-3 :  {A : Set} → {m n : ℕ} → {x : A} →  (lm : Vec A m) → (ln : Vec A n) →(e : n ≡ m) →  lm ≡ coe (vect-eq-1 e) ln → ( x ∷ lm) ≡ coe (vect-eq-1 (cong suc e))( x ∷ ln)
vect-eq-3 lm ln refl refl = refl



concat-assoc :  {A : Set} → (m n p : ℕ) → (lm : Vec A m) → (ln : Vec A n) → ( lp : Vec A p) → concatenation lm (concatenation ln lp) ≡  coe (vect-eq-1 (+-assoc m n p))( concatenation (concatenation lm ln) lp)
concat-assoc m n p [] ln lp = refl
concat-assoc (suc m) n p (x ∷ lm) ln lp = vect-eq-3 (concatenation lm (concatenation ln lp)) (concatenation (concatenation lm ln) lp) (+-assoc m n p) (concat-assoc m n p lm ln lp)

import Relation.Binary.HeterogeneousEquality

data _≅_ : {A B : Set} (x : A) (y : B) → Set where
  refl : {A : Set} {x : A} → x ≅ x

vect-eq :  {A : Set} → {m n : ℕ} → {x : A} →  (lm : Vec A m) → (ln : Vec A n) →  m ≡ n →  lm  ≅ ln →  (x ∷ lm) ≅ (x ∷ ln)
vect-eq lm .lm refl refl = refl

concat-assoc-easy :  {A : Set} → (m n p : ℕ) → (lm : Vec A m) → (ln : Vec A n) → ( lp : Vec A p) → concatenation lm (concatenation ln lp)  ≅   concatenation (concatenation lm ln) lp
concat-assoc-easy m n p [] ln lp = refl
concat-assoc-easy (suc m) n p (x ∷ lm) ln lp = vect-eq (concatenation lm (concatenation ln lp)) (concatenation (concatenation lm ln) lp) (sym (+-assoc m n p)) (concat-assoc-easy m n p lm ln lp)

≅-∷ : {A : Set} → (n m : ℕ) → (x : A) → (a : Vec A n) → (b : Vec A m) → (n ≡ m) → ( a ≅ b ) → ((x ∷ a) ≅ (x ∷ b) )
≅-∷ n m x a b refl refl = refl
